/*
 *  Mesh.cpp
 *  alife
 *
 *  Created by Eduardo Poyart on 3/18/10.
 *  Copyright 2010 __MyCompanyName__. All rights reserved.
 *
 */

#include "alife_Prefix.pch"
#include "common.h"

static bool is_power_of_2(int x) { return (x & (x - 1)) == 0; } 

void Mesh::init(dSpaceID in_space)
{
	assert(is_power_of_2(m_nx - 1));
	assert(is_power_of_2(m_ny - 1));
	for (int j = 0; j < m_ny; j++)
		for (int i = 0; i < m_nx; i++)
			m_hmap(i, j) = -0.4 * (real)i;

	int nbits = 0;
	int bx = m_nx;
	int by = m_ny;
	while (bx > 0 || by > 0)
	{
		nbits++;
		bx >>= 1;
		by >>= 1;
	}
	
	srand(123008);
	static const real rand_factor = 0.4;
	
	for (int k = 0; k < nbits; k++)
	{
		int step = 1 << (nbits-k-1);
		for (int j = 0; j < m_ny; j += step)
			for (int i = 0; i < m_nx; i += step)
				m_hmap(i, j) += (((real)rand() / RAND_MAX) - 0.5) * rand_factor;
		interpolate(step);
	}
	create_ode_trimesh(in_space);
}

void Mesh::interpolate_row(int i0, int i1, int j)
{
	real s = i1 - i0;
	real v0 = m_hmap(i0, j);
	real v1 = m_hmap(i1, j);
	for (int i = i0 + 1; i < i1; i++)
	{
		real r = (i - i0) / s;
		m_hmap(i, j) = v0 * (1.0 - r) + v1 * r;
	}
}

void Mesh::interpolate_col(int i, int j0, int j1)
{
	real s = j1 - j0;
	real v0 = m_hmap(i, j0);
	real v1 = m_hmap(i, j1);
	for (int j = j0 + 1; j < j1; j++)
	{
		real r = (j - j0) / s;
		m_hmap(i, j) = v0 * (1.0 - r) + v1 * r;
	}
}

void Mesh::interpolate(int step)
{
	for (int j = 0; j < m_ny-1; j += step)
		for (int i = 0; i < m_nx-1; i += step)
			interpolate_row(i, i+step, j);
	for (int j = 0; j < m_ny-1; j += step)
		for (int i = 0; i < m_nx-1; i++)
			interpolate_col(i, j, j+step);
}

void Mesh::create_ode_trimesh(dSpaceID in_space)
{
	m_trimesh = dGeomTriMeshDataCreate();
	m_num_indexes = (m_nx - 1) * (m_ny - 1) * 2 * 3;
	m_indexes = new dTriIndex[m_num_indexes];
	int i = 0;
	for (int y = 0; y < m_ny - 1; y++) 
	{
		for (int x = 0; x < m_nx - 1; x++) 
		{
			int v = y * m_nx + x;
			assert(i < m_num_indexes);
			m_indexes[i++] = v;
			m_indexes[i++] = v + 1;
			m_indexes[i++] = v + m_nx;
			m_indexes[i++] = v + m_nx;
			m_indexes[i++] = v + 1;
			m_indexes[i++] = v + m_nx + 1;
		}
	}
	assert(i == m_num_indexes);
	i = 0;
	m_num_verts = m_nx * m_ny;
	m_verts = new dReal[m_num_verts * 3];
	Vector3 offset(-2.0, -m_spacing * 0.3 * m_ny, 0.1);
	for (int y = 0; y < m_ny; y++)
	{
		for (int x = 0; x < m_nx; x++)
		{
			m_verts[i++] = x * m_spacing + offset.x;
			m_verts[i++] = y * m_spacing + offset.y;
			m_verts[i++] = m_hmap(x, y) + offset.z;
			
		}
	}
	assert(i == m_num_verts * 3);
	m_normals = new real[m_num_verts * 3];
	for (i = 0; i < m_num_verts * 3; i += 3)
	{
		m_normals[i] = 0.0;
		m_normals[i+1] = 0.0;
		m_normals[i+2] = 1.0;
	}
	Vector3 v;
	Vector3 n;
	Vector3 vdir[4];
	Vector3 vn[4];
	for (int y = 1; y < m_ny - 1; y++)
	{
		for (int x = 1; x < m_nx - 1; x++)
		{
			i = (y * m_nx + x) * 3;
			v = Vector3(&m_verts[i]);
			vdir[0] = (Vector3(&m_verts[i+3])-v).normalize();
			vdir[1] = (Vector3(&m_verts[i-3*m_nx])-v).normalize();
			vdir[2] = (Vector3(&m_verts[i-3])-v).normalize();
			vdir[3] = (Vector3(&m_verts[i+3*m_nx])-v).normalize();
			/*
			vdir[0] = (Vector3(&m_verts[i+3])-v);
			vdir[1] = (Vector3(&m_verts[i-3*m_nx])-v);
			vdir[2] = (Vector3(&m_verts[i-3])-v);
			vdir[3] = (Vector3(&m_verts[i+3*m_nx])-v);
			 */
			for (int j = 0; j < 4; j++)
				vdir[j].z *= 2.0;
			for (int j = 0; j < 4; j++)
				vn[j] = cross(vdir[(j+1)%4], vdir[j]);
			n = (vn[0]+vn[1]+vn[2]+vn[3]) * 0.25;
			n.normalize();
			m_normals[i] = n.x;
			m_normals[i+1] = n.y;
			m_normals[i+2] = n.z;
		}
	}
	
	dGeomTriMeshDataBuildSingle(m_trimesh, m_verts, 3 * sizeof(float), m_num_verts, m_indexes, m_num_indexes, 3 * sizeof(dTriIndex));	
	m_geom = dCreateTriMesh(in_space, m_trimesh, NULL, NULL, NULL);
}

void Mesh::draw()
{
#if RENDER
	glShadeModel(GL_SMOOTH);
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	glVertexPointer(3, GL_FLOAT, 0, m_verts);
	glNormalPointer(GL_FLOAT, 0, m_normals);
	glDrawElements(GL_TRIANGLES, m_num_indexes, GL_UNSIGNED_INT, m_indexes);
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
	glShadeModel(GL_FLAT);
#endif
}
