/*
 *  Model.cpp
 *  alife
 *
 *  Created by Eduardo Poyart on 2/17/10.
 *  Copyright 2010 __MyCompanyName__. All rights reserved.
 *
 */

#include "alife_Prefix.pch"
#include "common.h"

void Model::create_geom(Solid& s, dSpaceID space)
{
	if (s.type == SOLID_BOX)
		s.geomid = dCreateBox(0, s.scale[0], s.scale[1], s.scale[2]);
	else if (s.type == SOLID_SPHERE)
		s.geomid = dCreateSphere(0, s.scale[0]*0.5);
	else if (s.type == SOLID_CYLINDER)
		s.geomid = dCreateCylinder(0, s.scale[0]*0.5, s.scale[2]);
	else
		assert(0);
	dGeomSetBody(s.geomid, s.bodyid);
	dGeomSetPosition(s.geomid, s.translation[0], s.translation[1], s.translation[2]);
	dGeomSetQuaternion(s.geomid, s.rotation.get_v());
	dSpaceAdd(space, s.geomid);	
}

void Model::add_geom(Solid& s, dSpaceID space)
{
	s.bodyid = 0;
	create_geom(s, space);
	m_geoms.push_back(s);
}

void Model::load(const string& filename, dWorldID world, dSpaceID space)
{
	ifstream ifs;
	ifs.open(filename.c_str(), ios::in);
	assert(ifs.good());
	
	Parser p(&ifs);
	while (!ifs.eof())
	{
		if (!p.find_token("Transform", false))
			break;
		load_solid(p);
	}
	ifs.close();
					
	for (int i = 0; i < m_solids.size(); i++)
	{
		Solid& s = m_solids[i];
		s.bodyid = dBodyCreate(world);
		dBodySetPosition(s.bodyid, s.translation[0], s.translation[1], s.translation[2]);
		dBodySetQuaternion(s.bodyid, s.rotation.get_v());
		dMass m;
		if (s.type == SOLID_BOX)
			dMassSetBox(&m, 1.0, s.scale[0], s.scale[1], s.scale[2]);
		else if (s.type == SOLID_SPHERE)
			dMassSetSphere(&m, 1.0, s.scale[0] * 0.5);
		else if (s.type == SOLID_CYLINDER)
			dMassSetCylinder(&m, 1.0, 3, s.scale[0] * 0.5, s.scale[2]);
		else
			assert(0);
		dMassAdjust(&m, 1.0);
		dBodySetMass(s.bodyid, &m);
		create_geom(s, space);
	}
	for (int i = 0; i < m_geoms.size(); i++)
	{
		Solid& s = m_geoms[i];
		s.bodyid = 0;
		create_geom(s, space);
	}
	
	for (int k = 0; k < 2; k++)
	{
		for (int i = 0; i < m_joints.size(); i++)
		{
			bool found = false;
			for (int j = 0; j < m_solids.size(); j++)
			{
				if (m_joints[i].bodyname[k] == m_solids[j].name)
				{
					m_joints[i].bodyid[k] = m_solids[j].bodyid;
					found = true;
				}
			}
			if (!found)
			{
				cout << "Error: Joint " << m_joints[i].name << " references non-existing body " << m_joints[i].bodyname[k] << endl;
				exit(1);
			}
		}
	}
	
	for (int i = 0; i < m_joints.size(); i++)
	{
		Joint& joint = m_joints[i];
		dMatrix3 r;
		dRfromQ(r, joint.rotation.get_v());
		Matrix33 rot(r[0], r[1], r[2], r[4], r[5], r[6], r[8], r[9], r[10]);
		Vector3 v1(rot * Vector3(1.0, 0.0, 0.0));
		Vector3 v2(rot * Vector3(0.0, 1.0, 0.0));
		if (joint.type == JOINT_UNIVERSAL)
		{
			dJointID j = dJointCreateUniversal(world, 0);
			joint.jointid = j;
			dJointAttach(j, joint.bodyid[0], joint.bodyid[1]);
			dJointSetUniversalAnchor(j, joint.translation[0], joint.translation[1], joint.translation[2]);
			dJointSetUniversalAxis1(j, v1[0], v1[1], v1[2]);
			dJointSetUniversalAxis2(j, v2[0], v2[1], v2[2]);
			if (joint.has_limit[0])
			{
				dJointSetUniversalParam(j, dParamLoStop, joint.limit[0].x);
				dJointSetUniversalParam(j, dParamHiStop, joint.limit[0].y);
				dJointSetUniversalParam(j, dParamLoStop, joint.limit[0].x);
			}
			if (joint.has_limit[1])
			{
				dJointSetUniversalParam(j, dParamLoStop2, joint.limit[1].x);
				dJointSetUniversalParam(j, dParamHiStop2, joint.limit[1].y);
				dJointSetUniversalParam(j, dParamLoStop2, joint.limit[1].x);
			}
		}
		else if (joint.type == JOINT_HINGE)
		{
			dJointID j = dJointCreateHinge(world, 0);
			joint.jointid = j;
			dJointAttach(j, joint.bodyid[0], joint.bodyid[1]);
			dJointSetHingeAnchor(j, joint.translation[0], joint.translation[1], joint.translation[2]);
			dJointSetHingeAxis(j, v1[0], v1[1], v1[2]);
			if (joint.has_limit[0])
			{
				dJointSetHingeParam(j, dParamLoStop, joint.limit[0].x);
				dJointSetHingeParam(j, dParamHiStop, joint.limit[0].y);
				dJointSetHingeParam(j, dParamLoStop, joint.limit[0].x);
			}			
		}
		else if (joint.type == JOINT_BALL)
		{
			dJointID j = dJointCreateBall(world, 0);
			joint.jointid = j;
			dJointAttach(j, joint.bodyid[0], joint.bodyid[1]);
			dJointSetBallAnchor(j, joint.translation[0], joint.translation[1], joint.translation[2]);
		}
		else
			assert(0);
	}

	m_head_bodyid = get_bodyid_by_name("Head");
	m_board_bodyid = 0;
	m_bicycle_bodyid = 0;

	if (g_params.model == MODEL_SNOWBOARDER)
	{
		m_nn = new NeuralNet(2, 1, 7, 7);
		m_nn_in.resize(2);
		m_nn_out.resize(7);
	}
	else
	{
		m_nn = new NeuralNet(2, 1, 4, 2);
		m_nn_in.resize(2);
		m_nn_out.resize(2);
	}

	m_fj = new FourierJoint(16);

	if (g_params.model == MODEL_SNOWBOARDER)
	{
		m_board_bodyid = get_bodyid_by_name("Board");
		//m_input_joints.push_back(get_joint_by_name("ConstUnT")); // 2
		//	m_input_joints.push_back(get_joint_by_name("ConstHiKL")); // 1
		//	m_input_joints.push_back(get_joint_by_name("ConstHiKR")); // 1
		
		m_input_bodies.push_back(get_bodyid_by_name("Torso2")); // 1
		m_input_bodies.push_back(get_bodyid_by_name("Head")); // 1
		
		//m_control_feet_joints.push_back(get_joint_by_name("ConstUnFL")); // 1
		//m_control_feet_joints.push_back(get_joint_by_name("ConstUnFR")); // 0
		m_control_joints.push_back(get_joint_by_name("ConstUnT")); // 2

		m_spring_joints.push_back(get_joint_by_name("ConstUnFL"));
		m_spring_joints.push_back(get_joint_by_name("ConstUnFR"));
		m_spring_joints.push_back(get_joint_by_name("ConstHiKR"));
		m_spring_joints.push_back(get_joint_by_name("ConstHiHR"));
		m_spring_joints.push_back(get_joint_by_name("ConstHiKL"));
		m_spring_joints.push_back(get_joint_by_name("ConstHiHL"));
	}
	else if (g_params.model == MODEL_SPIDER)
	{
		m_control_joints.push_back(get_joint_by_name("ConstUnL1")); // 2
		m_control_joints.push_back(get_joint_by_name("ConstUnL2")); // 2
		m_control_joints.push_back(get_joint_by_name("ConstUnL3")); // 2
		m_control_joints.push_back(get_joint_by_name("ConstUnL4")); // 2
		m_control_joints.push_back(get_joint_by_name("ConstUnR1")); // 2
		m_control_joints.push_back(get_joint_by_name("ConstUnR2")); // 2
		m_control_joints.push_back(get_joint_by_name("ConstUnR3")); // 2
		m_control_joints.push_back(get_joint_by_name("ConstUnR4")); // 2
	}
	else if (g_params.model == MODEL_BOARD)
	{
		m_board_bodyid = get_bodyid_by_name("Board");
	}
	else if (g_params.model == MODEL_CYCLIST)
	{
		m_bicycle_bodyid = get_bodyid_by_name("BikeBody");
		m_bicycle_wheel1id = get_bodyid_by_name("WheelFront");
		m_bicycle_wheel2id = get_bodyid_by_name("WheelBack");
		m_control_joints.push_back(get_joint_by_name("ConstHiArmL")); // 1
		m_control_joints.push_back(get_joint_by_name("ConstHiArmR")); // 1	
	}
	else
		assert(0);
#if RENDER
	m_sphereobj = gluNewQuadric();
#endif
}

void Model::load_solid(Parser& p)
{
	string name;
	Vector3 translation;
	Vector3 scale;
	Vector4 rotation;
	string body1;
	string body2;
	bool is_joint = false;
	
	p.find_token("DEF");
	p.find_token("=");
	p.find_token("\"");
	name = p.get_token();
	if (name.substr(0, 5) == "Const")
		is_joint = true;

	p.find_token("translation");
	p.find_token("=");
	p.find_token("\"");
	for (int i = 0; i < 3; i++)
		translation[i] = atof(p.get_token().c_str());

	p.find_token("scale");
	p.find_token("=");
	p.find_token("\"");
	for (int i = 0; i < 3; i++)
		scale[i] = 2.0 * atof(p.get_token().c_str());
	
	p.find_token("rotation");
	p.find_token("=");
	p.find_token("\"");
	Vector3 axis;
	real angle;
	for (int i = 0; i < 3; i++)
		axis[i] = atof(p.get_token().c_str());
	angle = atof(p.get_token().c_str()) * 0.5;
	axis.normalize();
	for (int i = 0; i < 3; i++)
		rotation[i+1] = axis[i] * sin(angle);
	rotation[0] = cos(angle);
	rotation.normalize();
	
	p.find_token("\"");
	p.find_token(">");
	p.find_token("<");
	string sprop = p.get_token();
	if (is_joint)
	{
		assert(sprop == "Properties");
		Joint j;
		j.name = name;
		if (name.substr(0, 7) == "ConstUn")
			j.type = JOINT_UNIVERSAL;
		else if (name.substr(0, 7) == "ConstHi")
			j.type = JOINT_HINGE;
		else if (name.substr(0, 7) == "ConstBa")
			j.type = JOINT_BALL;
		else
		{
			cout << "Error: bad joint type " << name << endl;
			exit(1);
		}
		j.translation = translation;
		j.rotation = rotation;
		
		load_joint(p, j);
		
		m_joints.push_back(j);
	}
	else
	{
		Solid s;
		s.name = name;
		s.translation = translation;
		s.scale = scale;
		s.rotation = rotation;
		while (1)
		{
			string str = p.get_token();
			if (str == "coordIndex")
			{
				p.find_token("=");
				p.find_token("\"");
				str = p.get_token();
				if (str == "0")
					s.type = SOLID_BOX;
				else if (str == "126")
					s.type = SOLID_SPHERE;
				else // 64
					s.type = SOLID_CYLINDER;
			}
			else if (str == "/IndexedFaceSet")
				break;
		}		
		if (s.name.substr(0, 5) == "World")
			m_geoms.push_back(s);
		else
			m_solids.push_back(s);
	}
}

void Model::load_joint(Parser& p, Joint& j)
{
	string sprop;
	string stemp;
	sprop = p.get_token();
	while (sprop == "name")
	{
		p.find_token("=");
		p.find_token("\"");
		sprop = p.get_token();
		if (sprop == "Const1")
		{
			p.find_token("value");
			p.find_token("=");
			p.find_token("\"");
			j.bodyname[0] = p.get_token();
			p.find_token("\"");
		}
		else if (sprop == "Const2")
		{
			p.find_token("value");
			p.find_token("=");
			p.find_token("\"");
			j.bodyname[1] = p.get_token();
			p.find_token("\"");
		}
		else if (sprop == "Limit1")
		{
			p.find_token("value");
			p.find_token("=");
			p.find_token("\"");
			j.limit[0].x = atof(p.get_token().c_str());
			stemp = p.get_token();
			if (stemp != ",")
			{
				cout << "Error in " << j.name << ": use ',' to separate limits" << endl;
				exit(1);
			}
			j.limit[0].y = atof(p.get_token().c_str());
			j.has_limit[0] = true;
			p.find_token("\"");
		}
		else if (sprop == "Limit2")
		{
			p.find_token("value");
			p.find_token("=");
			p.find_token("\"");
			j.limit[1].x = atof(p.get_token().c_str());
			stemp = p.get_token();
			if (stemp != ",")
			{
				cout << "Error in " << j.name << ": use ',' to separate limits" << endl;
				exit(1);
			}
			j.limit[1].y = atof(p.get_token().c_str());
			j.has_limit[1] = true;
			p.find_token("\"");
		}
		else
		{
			cout << "Error: bad key " << sprop << endl;
			exit(1);
		}
		sprop = p.get_token();
	}
	if (sprop != "/")
	{
		cout << "Error: bad xml" << endl;
		exit(1);
	}
}

void Model::step(real t)
{
#if PROOF_OF_CONCEPT
	static real t_total = 0.0;
	t_total += t;
	for (int i = 0; i < m_control_joints.size(); i++)
	{
		Joint& j = m_joints[m_control_joints[i]];
		if (j.type == JOINT_UNIVERSAL)
		{
			//cout << "i = " << i << " ";
			set_pd_universal(j, (real)sin(t_total*4.0+(i*g_pi/2.0)), (real)cos(t_total*4.0+(i*g_pi/2.0)), 16.0, 0.8);
		}
		else if (j.type == JOINT_HINGE)
			set_pd_hinge(j, (real)sin(t_total*4.0+(i*g_pi)), 16.0, 1.0);
		else
			assert(0);
	}
#else
	if (g_params.model == MODEL_SNOWBOARDER)
		step_snowboarder(t);
	else if (g_params.model == MODEL_SPIDER)
		step_spider(t);
	else if (g_params.model == MODEL_BOARD) {}
	else if (g_params.model == MODEL_CYCLIST)
		step_cyclist(t);
	else
		assert(0);
#endif
}

void Model::set_pd_universal(const Joint& jnt, real val1, real val2, real ks, real kd)
{
	//cout << "in: " << val1;
	assert(jnt.type == JOINT_UNIVERSAL);
	real curr_angle[2];
	real curr_vel[2];
	real vel;
	curr_angle[0] = dJointGetUniversalAngle1(jnt.jointid);
	curr_angle[1] = dJointGetUniversalAngle2(jnt.jointid);
	curr_vel[0] = dJointGetUniversalAngle1Rate(jnt.jointid);
	curr_vel[1] = dJointGetUniversalAngle2Rate(jnt.jointid);
	assert(jnt.has_limit[0] && jnt.has_limit[1]);
	real limit_delta = jnt.limit[0].y - jnt.limit[0].x;
	real limit_center = (jnt.limit[0].y + jnt.limit[0].x) * 0.5;
	real nval1 = (val1 * limit_delta * 0.5) + limit_center;
	assert(nval1 >= jnt.limit[0].x && nval1 <= jnt.limit[0].y);
	vel = ks * (nval1 - curr_angle[0]) - kd * curr_vel[0];
	//cout << " out: " << vel << endl;
	dJointSetUniversalParam(jnt.jointid, dParamVel, vel);
	dJointSetUniversalParam(jnt.jointid, dParamFMax, ks * 5.0);
	limit_delta = jnt.limit[1].y - jnt.limit[1].x;
	limit_center = (jnt.limit[1].y + jnt.limit[1].x) * 0.5;
	real nval2 = (val2 * limit_delta * 0.5) + limit_center;
	assert(nval2 >= jnt.limit[1].x && nval2 <= jnt.limit[1].y);
	vel = ks * (nval2 - curr_angle[1]) - kd * curr_vel[1];
	//cout << vel << " ";
	dJointSetUniversalParam(jnt.jointid, dParamVel2, vel);
	dJointSetUniversalParam(jnt.jointid, dParamFMax2, ks * 5.0);
}

void Model::set_pd_hinge(const Joint& jnt, real val, real ks, real kd)
{
	assert(jnt.type == JOINT_HINGE);
	real curr_angle = dJointGetHingeAngle(jnt.jointid);
	real curr_vel = dJointGetHingeAngleRate(jnt.jointid);
	assert(jnt.has_limit[0]);
	real limit_delta = jnt.limit[0].y - jnt.limit[0].x;
	real limit_center = (jnt.limit[0].y + jnt.limit[0].x) * 0.5;
	real nval = (val * limit_delta * 0.5) + limit_center;
	real vel = ks * (nval - curr_angle) - kd * curr_vel;
	//cout << vel << " ";
	dJointSetHingeParam(jnt.jointid, dParamVel, vel);
	dJointSetHingeParam(jnt.jointid, dParamFMax, ks * 5.0);
}

void Model::set_spring_universal(const Joint& jnt, real val1, real val2, real ks, real kd)
{
	assert(jnt.type == JOINT_UNIVERSAL);
	real curr_angle[2];
	real curr_vel[2];
	real force1, force2;
	curr_angle[0] = dJointGetUniversalAngle1(jnt.jointid);
	curr_angle[1] = dJointGetUniversalAngle2(jnt.jointid);
	curr_vel[0] = dJointGetUniversalAngle1Rate(jnt.jointid);
	curr_vel[1] = dJointGetUniversalAngle2Rate(jnt.jointid);
	/*
	assert(jnt.has_limit[0] && jnt.has_limit[1]);
	real limit_delta = jnt.limit[0].y - jnt.limit[0].x;
	real limit_center = (jnt.limit[0].y + jnt.limit[0].x) * 0.5;
	real nval1 = (val1 * limit_delta * 0.5) + limit_center;
	assert(nval1 >= jnt.limit[0].x && nval1 <= jnt.limit[0].y);
	vel = ks * (nval1 - curr_angle[0]) - kd * curr_vel[0];
	dJointSetUniversalParam(jnt.jointid, dParamVel, vel);
	dJointSetUniversalParam(jnt.jointid, dParamFMax, ks * 5.0);
	limit_delta = jnt.limit[1].y - jnt.limit[1].x;
	limit_center = (jnt.limit[1].y + jnt.limit[1].x) * 0.5;
	real nval2 = (val2 * limit_delta * 0.5) + limit_center;
	assert(nval2 >= jnt.limit[1].x && nval2 <= jnt.limit[1].y);
	vel = ks * (nval2 - curr_angle[1]) - kd * curr_vel[1];
	dJointSetUniversalParam(jnt.jointid, dParamVel2, vel);
	dJointSetUniversalParam(jnt.jointid, dParamFMax2, ks * 5.0);
	*/
	force1 = ks * (val1 - curr_angle[0]) - kd * curr_vel[0];
	force2 = ks * (val2 - curr_angle[1]) - kd * curr_vel[1];
	dJointAddUniversalTorques(jnt.jointid, force1, force2);
}

void Model::set_spring_hinge(const Joint& jnt, real val, real ks, real kd)
{
	assert(jnt.type == JOINT_HINGE);
	real curr_angle = dJointGetHingeAngle(jnt.jointid);
	real curr_vel = dJointGetHingeAngleRate(jnt.jointid);
	/*
	assert(jnt.has_limit[0]);
	real limit_delta = jnt.limit[0].y - jnt.limit[0].x;
	real limit_center = (jnt.limit[0].y + jnt.limit[0].x) * 0.5;
	real nval = (val * limit_delta * 0.5) + limit_center;
	real vel = ks * (nval - curr_angle) - kd * curr_vel;
	dJointSetHingeParam(jnt.jointid, dParamVel, vel);
	dJointSetHingeParam(jnt.jointid, dParamFMax, ks * 5.0);
	*/
	real force = ks * (val - curr_angle) - kd * curr_vel;
	dJointAddHingeTorque(jnt.jointid, force);
}

#define INPUT_FACTOR 10.0

void Model::step_snowboarder(real t)
{
	int c = 0;
	for (int i = 0; i < m_input_joints.size(); i++)
	{
		Joint& j = m_joints[m_input_joints[i]];
		if (j.type == JOINT_UNIVERSAL)
		{
			m_nn_in[c++] = dJointGetUniversalAngle1(j.jointid) * INPUT_FACTOR;
			m_nn_in[c++] = dJointGetUniversalAngle2(j.jointid) * INPUT_FACTOR;
		}
		else
			m_nn_in[c++] = dJointGetHingeAngle(j.jointid) * INPUT_FACTOR;
	}
	dBodyID boardid = get_board_bodyid();
	const dReal* pos0 = dBodyGetPosition(boardid);
	const dReal* rot = dBodyGetRotation(boardid);
	Vector3 board_dir(rot[1], rot[5], 0.0);
	board_dir.normalize();
	for (int i = 0; i < m_input_bodies.size(); i++)
	{
		Vector3 pos(dBodyGetPosition(m_input_bodies[i]));
		real proj = dot(pos - pos0, board_dir);
		m_nn_in[c++] = proj * INPUT_FACTOR;
	}
	assert(c == 2);
	
	m_nn->compute(m_nn_in, m_nn_out);
	
	/*
	for (int i = 0; i < 3; i++)
		cout << m_nn_out[i] << " ";
	cout << endl;
	 */
	
	real ks = 8.0;
	real kd = ks * 0.1;
	real spring_ks = 8.0;
	real spring_kd = spring_ks * 0.1;

	c = 0;
	real feet_val;
	if (m_control_feet_joints.size() > 0)
		feet_val = m_nn_out[c++];
	for (int i = 0; i < m_control_feet_joints.size(); i++)
	{
		Joint& j = m_joints[m_control_feet_joints[i]];
		set_pd_universal(j, feet_val, 0.0, ks, kd);
	}
	for (int i = 0; i < m_control_joints.size(); i++)
	{
		Joint& j = m_joints[m_control_joints[i]];
		if (j.type == JOINT_HINGE)
			set_pd_hinge(j, m_nn_out[c++], ks, kd);
		else
			set_pd_universal(j, m_nn_out[c++], m_nn_out[c++], ks, kd);
	}
	assert(c == 2);
	for (int i = 0; i < m_spring_joints.size(); i++)
	{
		Joint& j = m_joints[m_spring_joints[i]];
		if (j.type == JOINT_HINGE)
			set_pd_hinge(j, 0.0, spring_ks, spring_kd);
		else
			set_pd_universal(j, 0.0, 0.0, spring_ks, spring_kd);
	}
}

void Model::step_cyclist(real t)
{
	int c = 0;
	dBodyID bikeid = get_bicycle_bodyid();
	const dReal* rot = dBodyGetRotation(bikeid);
	Vector3 bike_side(rot[1], rot[5], 0.0);
	bike_side.normalize();
	Vector3 bike_proj(rot[2], rot[6], 0.0);
	real bike_angle = dot(bike_proj, bike_side);
	//cout << bike_angle << endl;
	m_nn_in[c++] = bike_angle * INPUT_FACTOR;
	const dReal* vel = dBodyGetLinearVel(bikeid);
	Vector3 bike_vel(vel[0], vel[1], vel[2]);
	real speed = bike_vel.length();
	bike_vel.normalize();
	m_nn_in[c++] = bike_angle * INPUT_FACTOR;
	assert(c == 2);
	real brake_force = 0.0;
	Vector3 brake_dir(-rot[0], -rot[4], -rot[8]);
	const real max_speed = 5.0;
	if (speed > max_speed)
		brake_force = (speed - max_speed) * 10.0;
	brake_dir *= brake_force;
	dBodyAddForce(bikeid, brake_dir[0], brake_dir[1], brake_dir[2]);
	
	m_nn->compute(m_nn_in, m_nn_out);
	
	real ks = 8.0;
	real kd = ks * 0.1;
	c = 0;	
	for (int i = 0; i < m_control_joints.size(); i++)
	{
		Joint& j = m_joints[m_control_joints[i]];
		if (j.type == JOINT_HINGE)
			set_pd_hinge(j, m_nn_out[c++], ks, kd);
		else
			set_pd_universal(j, m_nn_out[c++], m_nn_out[c++], ks, kd);
	}
	assert(c == 2);
}


static const real ks = 5.0;
static const real kd = ks * 0.05;

void Model::step_spider(real t)
{
	m_total_time += t;
	for (int i = 0; i < m_control_joints.size(); i++)
	{
		Joint& jnt = m_joints[m_control_joints[i]];
		assert(jnt.type == JOINT_UNIVERSAL);
		real curr_angle[2];
		real curr_vel[2];
		real vel[2];
		curr_angle[0] = dJointGetUniversalAngle1(jnt.jointid);
		curr_angle[1] = dJointGetUniversalAngle2(jnt.jointid);
		curr_vel[1] = dJointGetUniversalAngle1Rate(jnt.jointid);
		curr_vel[2] = dJointGetUniversalAngle2Rate(jnt.jointid);
		for (int j = 0; j < 2; j++)
		{
			real torque = m_fj->get_torque(2*i+j, m_total_time);
			vel[j] = torque - curr_angle[j];
			vel[j] *= ks;
			vel[j] -= kd * curr_vel[j];
		};
		dJointSetUniversalParam(jnt.jointid, dParamVel, vel[0]);
		dJointSetUniversalParam(jnt.jointid, dParamFMax, 10.0);
		dJointSetUniversalParam(jnt.jointid, dParamVel2, vel[1]);
		dJointSetUniversalParam(jnt.jointid, dParamFMax2, 10.0);
	}
}

static void set_transform(const real pos[3], const real R[12])
{
#if RENDER
	GLfloat matrix[16];
	matrix[0]=R[0];
	matrix[1]=R[4];
	matrix[2]=R[8];
	matrix[3]=0;
	matrix[4]=R[1];
	matrix[5]=R[5];
	matrix[6]=R[9];
	matrix[7]=0;
	matrix[8]=R[2];
	matrix[9]=R[6];
	matrix[10]=R[10];
	matrix[11]=0;
	matrix[12]=pos[0];
	matrix[13]=pos[1];
	matrix[14]=pos[2];
	matrix[15]=1;
	glMultMatrixf(matrix);
#endif
}

void Model::draw_box(const real sides[3]) const
{
#if RENDER
	real lx = sides[0]*0.5f;
	real ly = sides[1]*0.5f;
	real lz = sides[2]*0.5f;
	
	// sides
	glBegin (GL_TRIANGLE_STRIP);
	glNormal3f (-1,0,0);
	glVertex3f (-lx,-ly,-lz);
	glVertex3f (-lx,-ly,lz);
	glVertex3f (-lx,ly,-lz);
	glVertex3f (-lx,ly,lz);
	glNormal3f (0,1,0);
	glVertex3f (lx,ly,-lz);
	glVertex3f (lx,ly,lz);
	glNormal3f (1,0,0);
	glVertex3f (lx,-ly,-lz);
	glVertex3f (lx,-ly,lz);
	glNormal3f (0,-1,0);
	glVertex3f (-lx,-ly,-lz);
	glVertex3f (-lx,-ly,lz);
	glEnd();
	
	// top face
	glBegin (GL_TRIANGLE_FAN);
	glNormal3f (0,0,1);
	glVertex3f (-lx,-ly,lz);
	glVertex3f (lx,-ly,lz);
	glVertex3f (lx,ly,lz);
	glVertex3f (-lx,ly,lz);
	glEnd();
	
	// bottom face
	glBegin (GL_TRIANGLE_FAN);
	glNormal3f (0,0,-1);
	glVertex3f (-lx,-ly,-lz);
	glVertex3f (-lx,ly,-lz);
	glVertex3f (lx,ly,-lz);
	glVertex3f (lx,-ly,-lz);
	glEnd();
#endif
}

// From drawstuff
// draw a cylinder of length l and radius r, aligned along the z axis
void Model::draw_cylinder(float l, float r, float zoffset) const
{
#if RENDER
	int i;
	float tmp,ny,nz,a,ca,sa;
	const int n = 24;	// number of sides to the cylinder (divisible by 4)
	
	l *= 0.5;
	r *= 0.5;
	a = float(M_PI*2.0)/float(n);
	sa = (float) sin(a);
	ca = (float) cos(a);
	
	// draw cylinder body
	ny=1; nz=0;		  // normal vector = (0,ny,nz)
	glBegin (GL_TRIANGLE_STRIP);
	for (i=0; i<=n; i++) {
		glNormal3d (ny,nz,0);
		glVertex3d (ny*r,nz*r,l+zoffset);
		glNormal3d (ny,nz,0);
		glVertex3d (ny*r,nz*r,-l+zoffset);
		// rotate ny,nz
		tmp = ca*ny - sa*nz;
		nz = sa*ny + ca*nz;
		ny = tmp;
	}
	glEnd();
	
	// draw top cap
	ny=1; nz=0;		  // normal vector = (0,ny,nz)
	glBegin (GL_TRIANGLE_FAN);
	glNormal3d (0,0,1);
	glVertex3d (0,0,l+zoffset);
	for (i=0; i<=n; i++) {
		if (i==1 || i==n/2+1)
			glColor3f(0.75f, 0.75f, 0.75f);
		//setColor (color[0]*0.75f,color[1]*0.75f,color[2]*0.75f,color[3]);
		glNormal3d (0,0,1);
		glVertex3d (ny*r,nz*r,l+zoffset);
		if (i==1 || i==n/2+1)
			glColor3f(1.0f, 1.0f, 1.0f);
		//setColor (color[0],color[1],color[2],color[3]);
		
		// rotate ny,nz
		tmp = ca*ny - sa*nz;
		nz = sa*ny + ca*nz;
		ny = tmp;
	}
	glEnd();
	
	// draw bottom cap
	ny=1; nz=0;		  // normal vector = (0,ny,nz)
	glBegin (GL_TRIANGLE_FAN);
	glNormal3d (0,0,-1);
	glVertex3d (0,0,-l+zoffset);
	for (i=0; i<=n; i++) {
		if (i==1 || i==n/2+1)
			glColor3f(0.75f, 0.75f, 0.75f);
		glNormal3d (0,0,-1);
		glVertex3d (ny*r,nz*r,-l+zoffset);
		if (i==1 || i==n/2+1)
			glColor3f(1.0f, 1.0f, 1.0f);
		
		// rotate ny,nz
		tmp = ca*ny + sa*nz;
		nz = -sa*ny + ca*nz;
		ny = tmp;
	}
	glEnd();
#endif
}

void Model::draw_sphere(const real scale[3]) const
{
#if RENDER
	gluSphere(m_sphereobj, scale[0]*0.5, 32, 64);
#endif
}

void Model::draw_solid(const Solid& s) const
{
	if (s.type == SOLID_BOX)
		draw_box(s.scale.get_v());
	else if (s.type == SOLID_CYLINDER)
		draw_cylinder(s.scale[2], s.scale[0], 0.0f);
	else if (s.type == SOLID_SPHERE)
		draw_sphere(s.scale.get_v());
}

void Model::draw() const
{
#if RENDER
	for (int i = 0; i < m_solids.size(); i++)
	{
		glPushMatrix();
		set_transform(dBodyGetPosition(m_solids[i].bodyid), dBodyGetRotation(m_solids[i].bodyid));
		draw_solid(m_solids[i]);
		glPopMatrix();
	}
	for (int i = 0; i < m_geoms.size(); i++)
	{
		glColor3f(0.3f, 0.5f, 0.5f);
		glPushMatrix();
		set_transform(dGeomGetPosition(m_geoms[i].geomid), dGeomGetRotation(m_geoms[i].geomid));
		draw_solid(m_geoms[i]);
		glPopMatrix();
	}
#endif
}

void Model::get_head_pos(real* pos) const
{
	/*
	Vector3 a(0.0, 0.0, 0.0);
	for (int i = 0; i < m_solids.size(); i++)
	{
		const dReal* p = dBodyGetPosition(m_solids[i].bodyid);
		a += Vector3(p);
	}
	pos[0] = a[0] / m_solids.size();
	pos[1] = a[1] / m_solids.size();
	pos[2] = a[2] / m_solids.size();
	 */
	const dReal* p = dBodyGetPosition(m_head_bodyid);
	pos[0] = p[0];
	pos[1] = p[1];
	pos[2] = p[2];
}

/*
void Model::set_pos(const real* pos)
{
	Vector3 pos0;
	get_head_pos(pos0.get_v());
	Vector3 delta(Vector3(pos) - pos0);
	Vector3 newpos;
	for (int i = 0; i < m_solids.size(); i++)
	{
		const dReal* pos1 = dBodyGetPosition(m_solids[i].bodyid);
		newpos = Vector3(pos1) + delta;
		dBodySetPosition(m_solids[i].bodyid, newpos.x, newpos.y, newpos.z);
		dGeomSetPosition(m_solids[i].geomid, newpos.x, newpos.y, newpos.z);
	}
}
 */

int Model::get_solid_index_by_name(const string& name)
{
	for (int i = 0; i < m_solids.size(); i++)
		if (m_solids[i].name == name)
			return i;
	assert(0);
	return -1;
}

int Model::get_joint_by_name(const string& name)
{
	for (int i = 0; i < m_joints.size(); i++)
		if (m_joints[i].name == name)
			return i;
	assert(0);
	return -1;
}

dBodyID Model::get_bodyid_by_name(const string& name)
{
	for (int i = 0; i < m_solids.size(); i++)
		if (m_solids[i].name == name)
			return m_solids[i].bodyid;
	assert(0);
	return 0;
}

dGeomID Model::get_geomid_by_name(const string& name)
{
	for (int i = 0; i < m_geoms.size(); i++)
		if (m_geoms[i].name == name)
			return m_geoms[i].geomid;
	assert(0);
	return 0;
}

void Model::replace_geomid_by_name(const string& name, dGeomID geom_id)
{
	int i;
	for (i = 0; i < m_geoms.size(); i++)
		if (m_geoms[i].name == name)
			break;
	assert(i < m_geoms.size());
	m_geoms[i].geomid = geom_id;
}

void Model::reset()
{
	m_total_time = 0;
	for (int i = 0; i < m_solids.size(); i++)
	{
		Solid& s = m_solids[i];
		dBodySetPosition(s.bodyid, s.translation[0], s.translation[1], s.translation[2]);
		dBodySetQuaternion(s.bodyid, s.rotation.get_v());
		dBodySetLinearVel(s.bodyid, 0.0, 0.0, 0.0);
		dBodySetAngularVel(s.bodyid, 0.0, 0.0, 0.0);
		dGeomSetPosition(s.geomid, s.translation[0], s.translation[1], s.translation[2]);
		dGeomSetQuaternion(s.geomid, s.rotation.get_v());
	}
}

