#include <math.h>
#include <memory.h>
#include <assert.h>
#include <stdio.h>
#define DISABLE_VECTOR_BOUNDS_CHECK
#include "base3d.h"
#undef DISABLE_VECTOR_BOUNDS_CHECK

const float g_pi        = 3.1415926536f;
const float g_halfpi    = 1.5707963268f;
const float g_twopi     = 6.2831853072f;
const float g_180overpi = 57.295779513f;

Vector3 spherical_to_cartesian( const Vector3 & v )
{
	// spherical: (r, phi, theta)
	// phi ~ y
	// theta ~ x or z
	// x = r sin phi cos theta
	// y = r cos phi
	// z = r sin phi sin theta
	float rsinphi = v.x * sinf(v.y);
	return Vector3(
		rsinphi * cosf(v.z),
		v.x * cosf(v.y),
		rsinphi * sinf(v.z) );
}

bool intersects( const Vector3 &p, const Vector3 &d,
                 const Vector3 &p1, const Vector3 &p2, const Vector3 &p3 )
{
	Vector3 e1( p2 - p1 ), e2( p3 - p1 ), h, s( p - p1 ), q;
	float a, f, u, v;

	h = cross( d, e2 );
	a = dot( e1, h );

	if ( a >= -EPSILON && a <= EPSILON )
		return false;

	f = 1.0f / a;
	u = f * dot( s, h );

	if ( u < 0.0f || u > 1.0f )
		return false ;

	q = cross( s, e1 );
	v = f * dot( d, q );
	if ( v < 0.0f || u + v > 1.0f )
		return false;

	return true;
}

float point_line_distance(const Vector3& line_pos, const Vector3& line_dir_unit,
                          const Vector3& point_pos, float& t)
{
	assert(line_dir_unit.length2() > 1.0f - EPSILON && line_dir_unit.length2() < 1.0f + EPSILON
		&& "point_line_distance: line_dir_unit must be normalized");
	Vector3 vecw = point_pos - line_pos;
	t = dot(vecw, line_dir_unit);
	return cross(line_dir_unit, vecw).length();
}

Vector2& Vector2::normalize()
{
	float invlength = 1.0f / sqrtf( x * x + y * y );
	x *= invlength;
	y *= invlength;
	return *this;
}

float Vector2::length() const
{
	return sqrtf( x * x + y * y );
}

float Vector2::invlength() const
{
	return 1.0f / sqrtf( x * x + y * y );
}

Vector3& Vector3::normalize()
{
	float invlength = 1.0f / sqrtf( x * x + y * y + z * z );
	x *= invlength;
	y *= invlength;
	z *= invlength;
	return *this;
}

float Vector3::length() const
{
	return sqrtf( x * x + y * y + z * z );
}

float Vector3::invlength() const
{
	return 1.0f / sqrtf( x * x + y * y + z * z );
}

Vector4& Vector4::normalize()
{
	float invlength = 1.0f / sqrtf( x * x + y * y + z * z + w * w );
	x *= invlength;
	y *= invlength;
	z *= invlength;
	w *= invlength;
	return *this;
}	

const Matrix33 Matrix33::Identity = Matrix33( 1.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 0.0f, 1.0f );

Matrix33::Matrix33(const Matrix33& m2)
{
	memcpy(&m[0][0], &m2.m[0][0], 9 * sizeof(float));
}

void Matrix33::createRotX( float t )
{
	float ct = cosf( t );
	float st = sinf( t );
	m[0] = Vector3( 1.0f, 0.0f, 0.0f );
	m[1] = Vector3( 0.0f, ct, -st );
	m[2] = Vector3( 0.0f, st, ct );
}

void Matrix33::createRotY( float t )
{
	float ct = cosf( t );
	float st = sinf( t );
	m[0] = Vector3( ct, 0.0f, st );
	m[1] = Vector3( 0.0f, 1.0f, 0.0f );
	m[2] = Vector3( -st, 0.0f, ct );
}

void Matrix33::createRotZ( float t )
{
	float ct = cosf( t );
	float st = sinf( t );
	m[0] = Vector3( ct, -st, 0.0f );
	m[1] = Vector3( st, ct, 0.0f );
	m[2] = Vector3( 0.0f, 0.0f, 1.0f );
}

const Matrix44 Matrix44::Identity = Matrix44( Matrix44::identity );

Matrix44::Matrix44(const Matrix44& m2)
{
	memcpy(&m[0][0], &m2.m[0][0], 16 * sizeof(float));
}

// From an OpenGL matrix
Matrix44::Matrix44(float* glm)
{
	memcpy(&m[0][0], glm, 16 * sizeof(float));
}

float Matrix44::determinant()
{
	Matrix33 m1(m[1][1], m[2][1], m[3][1],
	            m[1][2], m[2][2], m[3][2],
	            m[1][3], m[2][3], m[3][3]);

	Matrix33 m2(m[0][1], m[2][1], m[3][1],
	            m[0][2], m[2][2], m[3][2],
	            m[0][3], m[2][3], m[3][3]);

	Matrix33 m3(m[0][1], m[1][1], m[3][1],
	            m[0][2], m[1][2], m[3][2],
	            m[0][3], m[1][3], m[3][3]);

	Matrix33 m4(m[0][1], m[1][1], m[2][1],
	            m[0][2], m[1][2], m[2][2],
	            m[0][3], m[1][3], m[2][3]);

	return m[0][0] * m1.determinant() - m[1][0] * m2.determinant() + m[2][0] * m3.determinant() - m[3][0] * m4.determinant();
}

/*
bool Matrix44::invert(Matrix44& result)
{
	float det = determinant();
	if (det < EPSILON && det > -EPSILON)
		return false;
	Matrix33 sub;
	float subdet;
	float factor = 1.0f;
	int row1, row2, row3;
	Vector4 resultv;
	for (int i = 0; i < 4; i++) // loop through rows
	{
		if (i == 0)
		{
			row1 = 1; row2 = 2; row3 = 3;
		}
		else if (i == 1)
		{
			row1 = 0; row2 = 2; row3 = 3;
		}
		else if (i == 2)
		{
			row1 = 0; row2 = 1; row3 = 3;
		}
		else if (i == 3)
		{
			row1 = 0; row2 = 1; row3 = 2;
		}

		// column 0
		sub.c1 = Vector3(c2[row1], c2[row2], c2[row3]);
		sub.c2 = Vector3(c3[row1], c3[row2], c3[row3]);
		sub.c3 = Vector3(c4[row1], c4[row2], c4[row3]);
		subdet = sub.determinant();
		resultv[0] = subdet / det * factor;
		factor = -factor;

		// column 1
		sub.c1 = Vector3(c1[row1], c1[row2], c1[row3]);
		//sub.c2 = Vector3(c3[row1], c3[row2], c3[row3]);
		//sub.c3 = Vector3(c4[row1], c4[row2], c4[row3]);
		subdet = sub.determinant();
		resultv[1] = subdet / det * factor;
		factor = -factor;

		// column 2
		//sub.c1 = Vector3(c1[row1], c1[row2], c1[row3]);
		sub.c2 = Vector3(c2[row1], c2[row2], c2[row3]);
		//sub.c3 = Vector3(c4[row1], c4[row2], c4[row3]);
		subdet = sub.determinant();
		resultv[2] = subdet / det * factor;
		factor = -factor;

		// column 3
		//sub.c1 = Vector3(c1[row1], c1[row2], c1[row3]);
		//sub.c2 = Vector3(c2[row1], c2[row2], c2[row3]);
		sub.c3 = Vector3(c3[row1], c3[row2], c3[row3]);
		subdet = sub.determinant();
		resultv[3] = subdet / det * factor;
		Vector4& mv = (i == 0 ? result.c1 : (i == 1 ? result.c2 : (i == 2 ? result.c3 : result.c4)));
		mv = resultv;
	}
	return true;
}
*/

bool epsilon_equals(float a, float b)
{
	return (a >= b - EPSILON) && (a <= b + EPSILON);
}

bool epsilon_equals(float a, float b, float epsilon)
{
	return (a >= b - epsilon) && (a <= b + epsilon);
}

bool epsilon_equals(const Vector3& a, const Vector3& b)
{
	return epsilon_equals(a[0], b[0]) && epsilon_equals(a[1], b[1]) && epsilon_equals(a[2], b[2]);
}

bool epsilon_equals(const Vector3& a, const Vector3& b, float epsilon)
{
	return epsilon_equals(a[0], b[0], epsilon) && epsilon_equals(a[1], b[1], epsilon) && epsilon_equals(a[2], b[2], epsilon);
}

void position_orientation_to_opengl(const Vector3& position, const Matrix33& orientation, float *out_opengl_matrix)
{
    assert(out_opengl_matrix);
    float (*m44)[4] = (float(*)[4])out_opengl_matrix;
    int i;

    for (i = 0; i < 3; i++)
    {
		m44[0][i] = orientation[0][i];
		m44[1][i] = orientation[1][i];
		m44[2][i] = orientation[2][i];
    }

    m44[0][3] = m44[1][3] = m44[2][3] = 0.0f; m44[3][3] = 1.0f;

    for (i = 0; i < 3; i++)
    {
        m44[3][i] = position[i];
    }
}
