#ifndef _BASE3D_H_INCLUDED_
#define _BASE3D_H_INCLUDED_

extern const float g_pi;
extern const float g_halfpi;
extern const float g_twopi;
extern const float g_180overpi;

#define EPSILON 0.00001f

class Triangle
{
public:
	int p1;
	int p2;
	int p3;
	Triangle( int _p1, int _p2, int _p3 ) : p1(_p1), p2(_p2), p3(_p3) { };
};

class Vector2
{
public:
	float x, y;
	Vector2() {}
	Vector2(float _x, float _y): x(_x), y(_y) { }
	Vector2(const Vector2& v2): x(v2.x), y(v2.y) { }
	Vector2(const float* f): x(f[0]), y(f[1]) { }
	inline float& operator[](int i) 
	{
#ifndef DISABLE_VECTOR_BOUNDS_CHECK
		assert(i >= 0 && i <= 1);
#endif
		return (&x)[i];
	}
	inline const float& operator[](int i) const
	{
#ifndef DISABLE_VECTOR_BOUNDS_CHECK
		assert(i >= 0 && i <= 1);
#endif
		return (&x)[i];
	}
	inline float* get_v() { return &x; }
	inline const float* get_v() const { return &x; }

	Vector2& normalize();
	inline Vector2& operator+= ( const Vector2& v2 )
	{
		x += v2.x; y += v2.y;
		return *this;
	}

	inline Vector2& operator-= ( const Vector2& v2 )
	{
		x -= v2.x; y -= v2.y;
		return *this;
	}

	inline Vector2& operator*= ( float f )
	{
		x *= f; y *= f;
		return *this;
	}

	inline Vector2 operator-() const
	{
		return Vector2(-x, -y);
	}

	inline float length2() const
	{
		return x * x + y * y;
	}

	float length() const;
	float invlength() const;	
};

inline Vector2 operator+ ( const Vector2& v1, const Vector2& v2 )
{
	return Vector2( v1.x + v2.x, v1.y + v2.y );
}

inline Vector2 operator- ( const Vector2& v1, const Vector2& v2 )
{
	return Vector2( v1.x - v2.x, v1.y - v2.y );
}

inline Vector2 operator* ( const Vector2& v, float f )
{
	return Vector2( v.x * f, v.y * f );
}

inline Vector2 operator* ( float f, const Vector2& v )
{
	return Vector2( v.x * f, v.y * f );
}

// -----------------------

class Vector3
{
public:
	float x, y, z;
	Vector3() { }
	Vector3(float _x, float _y, float _z): x(_x), y(_y), z(_z) { }
	Vector3(const Vector3& v2): x(v2.x), y(v2.y), z(v2.z) { }
	Vector3(const float* f): x(f[0]), y(f[1]), z(f[2]) { }
	inline float& operator[](int i) 
	{
#ifndef DISABLE_VECTOR_BOUNDS_CHECK
		assert(i >= 0 && i <= 2);
#endif
		return (&x)[i];
	}
	inline const float& operator[](int i) const
	{
#ifndef DISABLE_VECTOR_BOUNDS_CHECK
		assert(i >= 0 && i <= 2);
#endif
		return (&x)[i];
	}
	inline float* get_v() { return &x; }
	inline const float* get_v() const { return &x; }

	Vector3& normalize();
	inline Vector3& operator+= ( const Vector3& v2 )
	{
		x += v2.x; y += v2.y; z += v2.z;
		return *this;
	}

	inline Vector3& operator-= ( const Vector3& v2 )
	{
		x -= v2.x; y -= v2.y; z -= v2.z;
		return *this;
	}

	inline Vector3& operator*= ( float f )
	{
		x *= f; y *= f; z *= f;
		return *this;
	}

	inline Vector3 operator-() const
	{
		return Vector3(-x, -y, -z);
	}

	inline float length2() const
	{
		return x * x + y * y + z * z;
	}

	float length() const;
	float invlength() const;
};

inline Vector3 operator+ ( const Vector3& v1, const Vector3& v2 )
{
	return Vector3( v1.x + v2.x, v1.y + v2.y, v1.z + v2.z );
}

inline Vector3 operator- ( const Vector3& v1, const Vector3& v2 )
{
	return Vector3( v1.x - v2.x, v1.y - v2.y, v1.z - v2.z );
}

inline Vector3 operator* ( const Vector3& v, float f )
{
	return Vector3( v.x * f, v.y * f, v.z * f );
}

inline Vector3 operator* ( float f, const Vector3& v )
{
	return Vector3( v.x * f, v.y * f, v.z * f );
}

// -----------------------

class Vector4
{
public:
	float x, y, z, w;
	Vector4() { }
	Vector4(float _x, float _y, float _z, float _w): x(_x), y(_y), z(_z), w(_w) { }
	Vector4(const Vector4& v2): x(v2.x), y(v2.y), z(v2.z), w(v2.w) { }
	Vector4(const float* f): x(f[0]), y(f[1]), z(f[2]), w(f[3]) { }
	inline float& operator[](int i) 
	{
#ifndef DISABLE_VECTOR_BOUNDS_CHECK
		assert(i >= 0 && i <= 3);
#endif
		return (&x)[i];
	}
	inline const float& operator[](int i) const
	{
#ifndef DISABLE_VECTOR_BOUNDS_CHECK
		assert(i >= 0 && i <= 3);
#endif
		return (&x)[i];
	}
	inline float* get_v() { return &x; }
	inline const float* get_v() const { return &x; }

	Vector4& normalize();
	inline Vector4& operator+= ( const Vector4& v2 )
	{
		x += v2.x; y += v2.y; z += v2.z; w += v2.w;
		return *this;
	}

	inline Vector4& operator-= ( const Vector4& v2 )
	{
		x -= v2.x; y -= v2.y; z -= v2.z; w = v2.w;
		return *this;
	}

	inline Vector4& operator*= ( float f )
	{
		x *= f; y *= f; z *= f; w *= f;
		return *this;
	}

	inline Vector4 operator-() const
	{
		return Vector4(-x, -y, -z, -w);
	}

	inline float length2() const
	{
		return x * x + y * y + z * z + w * w;
	}

	float length() const;
	float invlength() const;
};

inline Vector4 operator+ ( const Vector4& v1, const Vector4& v2 )
{
	return Vector4( v1.x + v2.x, v1.y + v2.y, v1.z + v2.z, v1.w + v2.w );
}

inline Vector4 operator- ( const Vector4& v1, const Vector4& v2 )
{
	return Vector4( v1.x - v2.x, v1.y - v2.y, v1.z - v2.z, v1.w - v2.w );
}

inline Vector4 operator* ( const Vector4& v, float f )
{
	return Vector4( v.x * f, v.y * f, v.z * f, v.w * f );
}

inline Vector4 operator* ( float f, const Vector4& v )
{
	return Vector4( v.x * f, v.y * f, v.z * f, v.w * f );
}

// -----------------------

inline float cross( const Vector2& u, const Vector2& v )
{
	return u.x * v.y - u.y * v.x;
}

inline Vector3 cross( const Vector3& u, const Vector3& v )
{
	return Vector3( u.y * v.z - u.z * v.y,
	                u.z * v.x - u.x * v.z,
	                u.x * v.y - u.y * v.x );
}

inline float dot( const Vector2& u, const Vector2& v )
{
	return u.x * v.x + u.y * v.y;
}

inline float dot( const Vector3& u, const Vector3& v )
{
	return u.x * v.x + u.y * v.y + u.z * v.z;
}

// -----------------------

class Matrix33
{
public:
	Vector3 m[3]; // column vectors
	/*
	 [  m[0][0]  m[1][0]  m[2][0]  ]
	 [  m[0][1]  m[1][1]  m[2][1]  ]
	 [  m[0][2]  m[1][2]  m[2][2]  ]
	*/
	 
	Matrix33() {}
	
	// abc, def, ghi are rows.
	Matrix33(float a, float b, float c, float d, float e, float f, float g, float h, float i)
	{
		m[0][0] = a; m[1][0] = b; m[2][0] = c;
		m[0][1] = d; m[1][1] = e; m[2][1] = f;
		m[0][2] = g; m[1][2] = h; m[2][2] = i;
	}
	Matrix33(const Matrix33& m2);

	inline Vector3& operator[](int i) { return m[i]; }
	inline const Vector3& operator[](int i) const { return m[i]; }

	static const Matrix33 Identity;

	void createRotX( float t );
	void createRotY( float t );
	void createRotZ( float t );
	void setSkewSymmetric( const Vector3& v )
	{
		m[0][0] =  0.0f;   m[0][1] =  v[2];   m[0][2] = -v[1];
    	m[1][0] = -v[2];   m[1][1] =  0.0f;   m[1][2] =  v[0];
    	m[2][0] =  v[1];   m[2][1] = -v[0];   m[2][2] =  0.0f;
	}

	inline Vector3 operator*(const Vector3& v) const
	{
		return Vector3( m[0][0] * v[0] + m[1][0] * v[1] + m[2][0] * v[2],
		                m[0][1] * v[0] + m[1][1] * v[1] + m[2][1] * v[2],
		                m[0][2] * v[0] + m[1][2] * v[1] + m[2][2] * v[2] );
	}

	inline float determinant()
	{
		return m[0][0] * ( m[1][1] * m[2][2] - m[2][1] * m[1][2])
		     - m[1][0] * ( m[0][1] * m[2][2] - m[2][1] * m[0][2])
		     + m[2][0] * ( m[0][1] * m[1][2] - m[1][1] * m[0][2]);
	}

	inline void transpose()
	{
		float temp;
		temp = m[1][0]; m[1][0] = m[0][1]; m[0][1] = temp;
		temp = m[2][0]; m[2][0] = m[0][2]; m[0][2] = temp;
		temp = m[2][1]; m[2][1] = m[1][2]; m[1][2] = temp;
	}

	inline void orthonormalize()
	{
		Vector3 X(m[0][0], m[1][0], m[2][0]);
    	Vector3 Y(m[0][1], m[1][1], m[2][1]);
    	Vector3 Z;

    	X.normalize();
    	Z = cross(X,Y);
    	Z.normalize();
    	Y = cross(Z,X);
    	Y.normalize();

    	m[0][0] = X[0]; m[1][0] = X[1]; m[2][0] = X[2];
		m[0][1] = Y[0]; m[1][1] = Y[1]; m[2][1] = Y[2];
		m[0][2] = Z[0]; m[1][2] = Z[1]; m[2][2] = Z[2];
	}
};

class Matrix44
{
public:
	Vector4 m[4]; // column vectors

	Matrix44() {}
	Matrix44(const Matrix44& m2);

	/*
	// abcd, efgh, ijkl, nopq are rows.
	Matrix44(float a, float b, float c, float d, float e, float f, float g, float h,
			 float i, float j, float k, float l, float n, float o, float p, float q)
	{
		m[0][0] = a; m[1][0] = b; m[2][0] = c; m[3][0] = d;
		m[0][1] = e; m[1][1] = f; m[2][1] = g; m[3][1] = h;
		m[0][2] = i; m[1][2] = j; m[2][2] = k; m[3][2] = l;
		m[0][3] = n; m[1][3] = o; m[2][3] = p; m[3][3] = q;
	}
	 */
	
	// From an OpenGL matrix
	Matrix44(float* glm);

	enum SetIdentity { identity };
	Matrix44(SetIdentity)
	{
		m[0][0] = 1.0f; m[1][0] = 0.0f; m[2][0] = 0.0f; m[3][0] = 0.0f;
		m[0][1] = 0.0f; m[1][1] = 1.0f; m[2][1] = 0.0f; m[3][1] = 0.0f;
		m[0][2] = 0.0f; m[1][2] = 0.0f; m[2][2] = 1.0f; m[3][2] = 0.0f;
		m[0][3] = 0.0f; m[1][3] = 0.0f; m[2][3] = 0.0f; m[3][3] = 1.0f;
	}

	inline Vector4& operator[](int i) { return m[i]; }
	inline const Vector4& operator[](int i) const { return m[i]; }

	static const Matrix44 Identity;

	void createRotX( float t );
	void createRotY( float t );
	void createRotZ( float t );

	inline Vector4 operator*(const Vector4& v) const
	{
		return Vector4( m[0][0] * v[0] + m[1][0] * v[1] + m[2][0] * v[2] + m[3][0] * v[3],
		                m[0][1] * v[0] + m[1][1] * v[1] + m[2][1] * v[2] + m[3][1] * v[3],
		                m[0][2] * v[0] + m[1][2] * v[1] + m[2][2] * v[2] + m[3][2] * v[3],
		                m[0][3] * v[0] + m[1][3] * v[1] + m[2][3] * v[2] + m[3][3] * v[3] );
	}

	inline Vector3 operator*(const Vector3& v) const
	{
		return Vector3( m[0][0] * v[0] + m[1][0] * v[1] + m[2][0] * v[2] + m[3][0],
		                m[0][1] * v[0] + m[1][1] * v[1] + m[2][1] * v[2] + m[3][1],
		                m[0][2] * v[0] + m[1][2] * v[1] + m[2][2] * v[2] + m[3][2] );
	}

	float determinant();

	inline void orthonormalize()
	{
		Vector3 X(m[0][0], m[1][0], m[2][0]);
    	Vector3 Y(m[0][1], m[1][1], m[2][1]);
    	Vector3 Z;

    	X.normalize();
    	Z = cross(X,Y);
    	Z.normalize();
    	Y = cross(Z,X);
    	Y.normalize();

    	m[0][0] = X[0]; m[1][0] = X[1]; m[2][0] = X[2];
		m[0][1] = Y[0]; m[1][1] = Y[1]; m[2][1] = Y[2];
		m[0][2] = Z[0]; m[1][2] = Z[1]; m[2][2] = Z[2];
	}

	inline void invert_rot_tran()
	{
		//assert(m[0][3] == 0.0f && m[1][3] == 0.0f && m[2][3] == 0.0f && m[3][3] == 1.0f);

		// create transposed
		Matrix33 t(m[0][0], m[0][1], m[0][2],
				   m[1][0], m[1][1], m[1][2],
				   m[2][0], m[2][1], m[2][2]);
		Vector3 v(-m[3][0], -m[3][1], -m[3][2]);
		v = t * v;
		m[0][0] = t[0][0]; m[1][0] = t[1][0]; m[2][0] = t[2][0]; m[3][0] = v[0];
		m[0][1] = t[0][1]; m[1][1] = t[1][1]; m[2][1] = t[2][1]; m[3][1] = v[1];
		m[0][2] = t[0][2]; m[1][2] = t[1][2]; m[2][2] = t[2][2]; m[3][2] = v[2];
	}

};

inline Matrix33 transpose( const Matrix33& m )
{
	return Matrix33(m[0][0], m[0][1], m[0][2], m[1][0], m[1][1], m[1][2], m[2][0], m[2][1], m[2][2]);
}

inline Matrix33 operator+( const Matrix33& a, const Matrix33& b )
{
	return Matrix33( a[0][0] + b[0][0], a[1][0] + b[1][0], a[2][0] + b[2][0],
					 a[0][1] + b[0][1], a[1][1] + b[1][1], a[2][1] + b[2][1],
					 a[0][2] + b[0][2], a[1][2] + b[1][2], a[2][2] + b[2][2] );
}

inline Matrix33 operator*( const Matrix33& a, const Matrix33& b )
{
	return Matrix33(a[0][0] * b[0][0] + a[1][0] * b[0][1] + a[2][0] * b[0][2],
					a[0][0] * b[1][0] + a[1][0] * b[1][1] + a[2][0] * b[1][2],
					a[0][0] * b[2][0] + a[1][0] * b[2][1] + a[2][0] * b[2][2],
					a[0][1] * b[0][0] + a[1][1] * b[0][1] + a[2][1] * b[0][2],
					a[0][1] * b[1][0] + a[1][1] * b[1][1] + a[2][1] * b[1][2],
					a[0][1] * b[2][0] + a[1][1] * b[2][1] + a[2][1] * b[2][2],
					a[0][2] * b[0][0] + a[1][2] * b[0][1] + a[2][2] * b[0][2],
					a[0][2] * b[1][0] + a[1][2] * b[1][1] + a[2][2] * b[1][2],
					a[0][2] * b[2][0] + a[1][2] * b[2][1] + a[2][2] * b[2][2] );
}

/*
inline Matrix44 operator*( const Matrix44& a, const Matrix44& b )
{
	return Matrix44(a[0][0] * b[0][0] + a[1][0] * b[0][1] + a[2][0] * b[0][2] + a[3][0] * b[0][3],
					a[0][0] * b[1][0] + a[1][0] * b[1][1] + a[2][0] * b[1][2] + a[3][0] * b[1][3],
					a[0][0] * b[2][0] + a[1][0] * b[2][1] + a[2][0] * b[2][2] + a[3][0] * b[2][3],
					a[0][0] * b[3][0] + a[1][0] * b[3][1] + a[2][0] * b[3][2] + a[3][0] * b[3][3],
					
					a[0][1] * b[0][0] + a[1][1] * b[0][1] + a[2][1] * b[0][2] + a[3][1] * b[0][3],
					a[0][1] * b[1][0] + a[1][1] * b[1][1] + a[2][1] * b[1][2] + a[3][1] * b[1][3],
					a[0][1] * b[2][0] + a[1][1] * b[2][1] + a[2][1] * b[2][2] + a[3][1] * b[2][3],
					a[0][1] * b[3][0] + a[1][1] * b[3][1] + a[2][1] * b[3][2] + a[3][1] * b[3][3],
					
					a[0][2] * b[0][0] + a[1][2] * b[0][1] + a[2][2] * b[0][2] + a[3][2] * b[0][3],
					a[0][2] * b[1][0] + a[1][2] * b[1][1] + a[2][2] * b[1][2] + a[3][2] * b[1][3],
					a[0][2] * b[2][0] + a[1][2] * b[2][1] + a[2][2] * b[2][2] + a[3][2] * b[2][3],
					a[0][2] * b[3][0] + a[1][2] * b[3][1] + a[2][2] * b[3][2] + a[3][2] * b[3][3],

					a[0][3] * b[0][0] + a[1][3] * b[0][1] + a[2][3] * b[0][2] + a[3][3] * b[0][3],
					a[0][3] * b[1][0] + a[1][3] * b[1][1] + a[2][3] * b[1][2] + a[3][3] * b[1][3],
					a[0][3] * b[2][0] + a[1][3] * b[2][1] + a[2][3] * b[2][2] + a[3][3] * b[2][3],
					a[0][3] * b[3][0] + a[1][3] * b[3][1] + a[2][3] * b[3][2] + a[3][3] * b[3][3]					
					);
}
*/

inline Matrix33 operator*( const Matrix33& m, float f )
{
	return Matrix33( f * m[0][0], f * m[1][0], f * m[2][0],
					 f * m[0][1], f * m[1][1], f * m[2][1],
					 f * m[0][2], f * m[1][2], f * m[2][2] );
}

inline Matrix33 operator* ( float f, const Matrix33& m )
{
	return Matrix33( f * m[0][0], f * m[1][0], f * m[2][0],
					 f * m[0][1], f * m[1][1], f * m[2][1],
					 f * m[0][2], f * m[1][2], f * m[2][2] );
}

bool intersects( const Vector3 &p, const Vector3 &d,
				 const Vector3 &p1, const Vector3 &p2, const Vector3 &p3 );

float point_line_distance( const Vector3& line_pos, const Vector3& line_dir,
						   const Vector3& point_pos, float& t);

Vector3 spherical_to_cartesian( const Vector3 & v );

bool epsilon_equals(float a, float b);
bool epsilon_equals(float a, float b, float epsilon);
bool epsilon_equals(const Vector3& a, const Vector3& b);
bool epsilon_equals(const Vector3& a, const Vector3& b, float epsilon);

void position_orientation_to_opengl(const Vector3& position, const Matrix33& orientation, float *out_opengl_matrix);

#endif
